/***************************************************************************
**
**  This file is part of GpCoreTools.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program. If not, see <http://www.gnu.org/licenses/>.
**
**  See http://www.geopsy.org for more information.
**
**  Created: 2009-07-01
**  Authors:
**    Marc Wathelet
**    Marc Wathelet (LGIT, Grenoble, France)
**
***************************************************************************/

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <execinfo.h>
#include <cxxabi.h>

#include "CoreApplication.h"
#include "ApplicationHelp.h"
#include "PackageInfo.h"
#include "File.h"
#include "Log.h"
#include "Trace.h"
#include "Leds.h"
#include "PThread.h"

namespace GpCoreTools {

/*!
  \class CoreApplication CoreApplication.h
  \brief Brief description of class still missing

  Full description of class still missing
*/

CoreApplication * CoreApplication::_self=0;

/*!
  Description of constructor still missing
*/
CoreApplication::CoreApplication(int & argc, char ** argv,
                                 ApplicationHelp * ( *help) (),
                                 bool reportBugs)
{
  PThread::init();
  _applicationName=argv[0];
  int index=_applicationName.find_last_of("/");
  if(index>-1) {
    _applicationName=_applicationName.substr(index+1, _applicationName.length()-index);
  }
  srand(time(NULL)); // init random generator

  if(!_self) {
    _self=this;
  } else {
    Log::write(0, "error: only one instance of ApplicationCore is allowed.");
    abort();
  }
  // Size of terminal
  _terminalCols=0;
  _terminalRows=0;

  // Check arguments
  int i, j=1;
  for(i=1; i<argc; i++) {
    const char * arg=argv[i];
    if(help && (strcmp(arg, "--help")==0 || strcmp(arg, "-help")==0 || strcmp(arg, "-h")==0)) {
      showHelp(i, argc, argv, help);
      ::exit(0);
    } else if(strcmp(arg, "--verbosity")==0 || strcmp(arg, "-verbosity")==0 || strcmp(arg, "-v")==0) {
      checkOptionArg(i, argc, argv, true);
      Log::setVerbosity(atoi(argv[i]));
    } else if(strcmp(arg, "--nobugreport")==0 || strcmp(arg, "-nobugreport")==0) {
      reportBugs=false;
    } else if(strcmp(arg, "--version")==0 || strcmp(arg, "-version")==0) {
      printf("Versions:\n%s", version().data());
      ::exit(0);
    } else if(strcmp(arg, "--app-version")==0 || strcmp(arg, "-app-version")==0) {
      printf("%s %s\n", _applicationName.data(), version(_applicationName.data()).data());
      ::exit(0);
    } else {
      argv[j++]=argv[i];
    }
  }
  if(j<argc) {
    argv[j]=0;
    argc=j;
  }

  _reportBugs=reportBugs;
  initInternalDebugger();
}

/*!
  Description of destructor still missing
*/
CoreApplication::~CoreApplication()
{
}

/*!
  In case of fork, recall this function in the child.
*/
void CoreApplication::initInternalDebugger()
{
  if(_reportBugs) {
    signal (SIGABRT, osSignal);
    signal (SIGSEGV, osSignal);
    signal (SIGPIPE, osSignal);
    signal (SIGILL, osSignal);
  }
}

/*!
  In case an exception occurs during normal execution
*/
void CoreApplication::osSignal(int sig)
{
  static pthread_mutex_t lock;
  if(pthread_mutex_trylock(&lock)!=0) {
    signal(sig, SIG_DFL);
    sleep(10000);
    fprintf(stderr, "EXCEPTION occured while processing exceptions.\n");
    fprintf(stderr, "No backtrace information can be provided\n");
    fprintf(stderr, "Eventually kill remaining instances:\n");
    fprintf(stderr, "  ps\n" );
    fprintf(stderr, "  pkill <application name>\n" );
    exit(2);
  }

  Log::write(0, "received SEGV or ABRT signal\n");
  // 20151110: demangle and backtrace_symbols does not work on alix
  //           work only with TRACE_ENABLED
#ifdef TRACE_ENABLED
  PThread::lockThreads();
  std::list< std::vector<const TraceStamp *> * > sList=PThread::stacks();
  std::list< std::vector<const TraceStamp *> * >::iterator it;
  for(it=sList.begin(); it!=sList.end(); it++) {
    Log::write(0, "---- thread\n");
    Trace::backTrace(*it);
  }
  PThread::unlockThreads();
#else
  void *frames[50];
  size_t size;
  char ** strings;
  size_t i;
  size=backtrace (frames, 50);
  strings=backtrace_symbols (frames, size);

  size_t functionLen=0;
  int status;
  // Demangle, see doc at http://gcc.gnu.org/onlinedocs/libstdc++/latest-doxygen/namespaceabi.html
  for(i=0; i < size; i++) {
    // get value between (...+
    char * mangledName=strchr(strings[i], '(' );
    if(mangledName) {
      char * mangledNameEnd=strchr(mangledName, '+' );
      if(mangledNameEnd) {
        char * mangledString=new char[mangledNameEnd-mangledName];
        strncpy(mangledString, mangledName+1, mangledNameEnd-mangledName-1);
        mangledString[mangledNameEnd-mangledName-1]='\0';
        char * functionName=abi::__cxa_demangle(mangledString, 0, &functionLen, &status);
        Log::writeNoStamp(0, functionName);
        delete [] mangledString;
      } else {
        Log::writeNoStamp(0, strings[i]);
      }
    } else {
      Log::writeNoStamp(0, strings[i]);
    }
    Log::writeNoStamp(0, "\n");
  }
  free (strings);
#endif
  signal (SIGABRT, 0);
  abort();
}

/*!
  Returns a complete description of all versions
*/
std::string CoreApplication::version()
{
  char buf[256];
  std::string str;
  for(std::list<PackageInfo>::iterator it=PackageInfo::list()->begin();it!=PackageInfo::list()->end(); it++ ) {
    sprintf(buf, "  %-20s%s", +it->package().data(), it->version().data());
    str+=buf;
    if(it->distribution().empty()) {
      str+="\n";
    } else {
      str+=" from ";
      str+=it->distribution();
      str+="\n";
    }
    str+="      (built on ";
    str+=it->versionTime();
    str+=")\n";
  }
  return str;
}

/*!
  Returns the version of \a item
*/
std::string CoreApplication::version(const char * item)
{
  for(std::list<PackageInfo>::iterator it=PackageInfo::list()->begin();it!=PackageInfo::list()->end(); it++ ) {
    if(it->package()==item) {
      return it->version();
    }
  }
  return "";
}

/*!
  Returns the copyright message
*/
std::string CoreApplication::authors()
{
  for(std::list<PackageInfo>::iterator it=PackageInfo::list()->begin();it!=PackageInfo::list()->end(); it++ ) {
    if(it->package()==CoreApplication::instance()->applicationName()) {
      return it->authors();
    }
  }
  return "";
}

void CoreApplication::showHelp(int& i, int argc, char ** argv, ApplicationHelp * ( *help) ())
{
  ApplicationHelp * h=help();
  setHelp(h);
  if(checkOptionArg(i, argc, argv, false)) {
    h->exec(argv[i]);
  } else {
    h->exec();
  }
  delete h;
}

bool compareStrings(const char * first, const char * second)
{
  return strcmp(first, second)<0;
}

/*!

*/
void CoreApplication::setHelp(ApplicationHelp * h)
{
  h->addGroup("Generic", "generic");
  std::list<const char *> sections=h->sections();
  sections.sort(compareStrings);
  std::string str="Show help about options. SECTION may be empty or: all, html";
  for(std::list<const char *>::iterator it=sections.begin();it!=sections.end();it++) {
    str+= ", ";
    str+=*it;
  }
  h->addOption("-help <SECTION>", str.data());
  h->addOption("-verbosity <LEVEL>", "Select level of output message (default=0).");
  h->addOption("-version", "Show version information");
  h->addOption("-app-version", "Show short version information");
  h->addOption("-nobugreport", "Do not generate bug report in case of error");
}

/*!
  Check if option argument at \a i is available. If not and if \a mandatory is true, the application exits
  with error code 2. Else, it returns true if an argument is available.

  This function can be used even without a CoreApplication running.
*/
bool CoreApplication::checkOptionArg(int& i, int argc, char ** argv, bool mandatory)
{
  i++;
  if(i<argc) {
    if(argv[i][0]=='-') {
      char * ptr=0;
      float f=strtod(argv[i], &ptr);
      UNUSED(f);
      if(ptr>argv[i]) return true;
    } else return true;
  }
  if(mandatory) {
    Log::write(0, "error: option %s requires an argument, see -help for details.\n", argv[i-1]);
    CoreApplication::exit(2);
  }
  return false;
}

std::string CoreApplication::getStdin()
{
  std::string str;
  int n=256;
  char * buf=new char[n];
  while(true) {
    if(File::readLine(buf, n, stdin)) {
      if(strcmp(buf,"END\n")==0) break;
      str+=buf;
    }
  }
  delete [] buf;
  return str;
}

int CoreApplication::terminalRows() const
{
  if(_terminalRows==0) {
    terminalSize();
  }
  return _terminalRows;
}

int CoreApplication::terminalCols() const
{
  if(_terminalCols==0) {
    terminalSize();
  }
  return _terminalCols;
}

/*!
  Get the terminal size

  Originaly copied from ps command written by Albert Cahalan (Copyright 1998-2002),
  file global.c, package procps-3.2.7

  Comments of the original function (set_screen_size(void)):

  The rules:
    1. Defaults are implementation-specific. (ioctl,termcap,guess)
    2. COLUMNS and LINES override the defaults. (standards compliance)
    3. Command line options override everything else.
    4. Actual output may be more if the above is too narrow.

  SysV tends to spew semi-wide output in all cases. The args
  will be limited to 64 or 80 characters, without regard to
  screen size. So lines of 120 to 160 chars are normal.
  Tough luck if you want more or less than that! HP-UX has a
  new "-x" option for 1024-char args in place of comm that
  we'll implement at some point.

  BSD tends to make a good effort, then fall back to 80 cols.
  Use "ww" to get infinity. This is nicer for "ps | less"
  and "watch ps". It can run faster too.
*/
void CoreApplication::terminalSize() const
{
/*
 * Try not to overflow the output buffer:
 *    32 pages for env+cmd
 *    64 kB pages on IA-64
 *    4 chars for "\377", or 1 when mangling to '?'  (ESC_STRETCH)
 *    plus some slack for other stuff
 * That is about 8.5 MB on IA-64, or 0.6 MB on i386
 *
 * Sadly, current kernels only supply one page of env/command data.
 * The buffer is now protected with a guard page, and via other means
 * to avoid hitting the guard page.
 */

/* output buffer size */
// since we mangle to '?' this is 1 (would be 4 for octal escapes)
#define ESC_STRETCH 1
#define OUTBUF_SIZE (2 * 64*1024 * ESC_STRETCH)
#ifdef Q_OS_WIN
  _terminalCols=80;
  _terminalRows=1000;
#else
  struct winsize ws;
  char *columns; /* Unix98 environment variable */
  char *lines;   /* Unix98 environment variable */

  do{
    int fd;
    if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws)!=-1 && ws.ws_col>0 && ws.ws_row>0) break;
    if(ioctl(STDERR_FILENO, TIOCGWINSZ, &ws)!=-1 && ws.ws_col>0 && ws.ws_row>0) break;
    if(ioctl(STDIN_FILENO,  TIOCGWINSZ, &ws)!=-1 && ws.ws_col>0 && ws.ws_row>0) break;
    fd=open("/dev/tty", O_NOCTTY|O_NONBLOCK|O_RDONLY);
    if(fd!=-1){
      int ret=ioctl(fd, TIOCGWINSZ, &ws);
      ::close(fd);
      if(ret!=-1 && ws.ws_col>0 && ws.ws_row>0) break;
    }
    // TODO: ought to do tgetnum("co") and tgetnum("li") here
    ws.ws_col=80;
    ws.ws_row=24;
  } while(0);
  _terminalCols=ws.ws_col;  // hmmm, NetBSD subtracts 1
  _terminalRows=ws.ws_row;
  //fprintf(stderr,"Terminal size %i*%i\n",_terminalCols,_terminalRows);

  // TODO: delete this line
  //if(!isatty(STDOUT_FILENO)) _terminalCols=OUTBUF_SIZE;

  columns=getenv("COLUMNS");
  if(columns && *columns){
    long t;
    char *endptr;
    t=strtol(columns, &endptr, 0);
    if(!*endptr && (t>0) && (t<(long)OUTBUF_SIZE)) _terminalCols=(int)t;
  }

  lines  =getenv("LINES");
  if(lines && *lines){
    long t;
    char *endptr;
    t=strtol(lines, &endptr, 0);
    if(!*endptr && (t>0) && (t<(long)OUTBUF_SIZE)) _terminalRows=(int)t;
  }

  if((_terminalCols<9) || (_terminalRows<2))
    fprintf(stderr,"Your %dx%d screen size is bogus. Expect trouble.\n",
      _terminalCols, _terminalRows
    );
#endif
}

void CoreApplication::exit(int exitCode)
{
  if(exitCode!=0) {
    // If led support on alix board, fast blinking of first led to indicate problems
    Leds::timer(0, 100, 100);
  }
  if(instance()) {
    instance()->close();
  }
  ::exit(exitCode);
}

} // namespace GpCoreTools
