/***************************************************************************
**
**  This file is part of GpCoreTools.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program. If not, see <http://www.gnu.org/licenses/>.
**
**  See http://www.geopsy.org for more information.
**
**  Created: 2011-06-29
**  Authors:
**    Marc Wathelet (ISTerre, Grenoble, France)
**
***************************************************************************/

#include <signal.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#include "ApplicationHelp.h"
#include "EventLoop.h"
#include "DaemonApplication.h"

namespace GpCoreTools {

  /*!
    \class DaemonApplication DaemonApplication.h
    \brief Brief description of class still missing

    Full description of class still missing
  */

  DaemonApplication * DaemonApplication::_self=0;

  /*!
    Description of constructor still missing
  */
  DaemonApplication::DaemonApplication(int & argc, char ** argv, ApplicationHelp * ( *help) (), bool reportBugs)
    : CoreApplication(argc, argv, 0, reportBugs)
  {
    assert(!_self);
    _self=this;

    _log=0;
    _doFork=true;
    _basePath="/var/log";
    bool openLog=true;

    // Check arguments
    int i, j=1;
    for(i=1; i<argc; i++) {
      const char * arg=argv[i];
      if(strcmp(arg, "-N")==0) {
        _doFork=false;
      } else if(strcmp(arg, "-b")==0) {
        checkOptionArg(i, argc, argv, true);
        _basePath=argv[i];
      } else if(strcmp(arg, "-l")==0) {
        checkOptionArg(i, argc, argv, true);
        _logFileName=argv[i];
      } else if(strcmp(arg, "-no-log")==0) {
        openLog=false;
      } else if(help && (strcmp(arg, "--help")==0 || strcmp(arg, "-help")==0 || strcmp(arg, "-h")==0)) {
        showHelp(i, argc, argv, help);
        ::exit(0);
      } else {
        argv[j++]=argv[i];
      }
    }
    if(j<argc) {
      argv[j]=0;
      argc=j;
    }

    if(_doFork) {
      checkPidFile();
    }
    checkBasePath();
    if(openLog) {
      initLogFile();
    } else {
      Log::setVerbosity(-1);
    }

    // Allocate it after initializing log
    _leds=new Leds;
  }

  DaemonApplication::~DaemonApplication()
  {
    delete _leds;
    delete _log;
  }

  /*!

  */
  void DaemonApplication::close()
  {
    Log::write(0, "stop daemon with pid %i\n", getpid());
    Log::write(0, "pid file %s removed\n", pidFile(applicationName()).data());
    remove(pidFile(applicationName()).data());
    CoreApplication::close();
  }

  std::string DaemonApplication::pidFile(const char * applicationName)
  {
    std::string f="/var/run/";
    f+=applicationName;
    f+=".pid";
    return f;
  }

  void DaemonApplication::checkPidFile()
  {
    struct stat sInfo;
    std::string pf=pidFile(applicationName());
    int sErr=stat(pf.data(), &sInfo);
    if(sErr==0) {
      fprintf(stderr, "%s is already running (stop it and remove %s)\n",
               applicationName(), pf.data());
      // Do not use CoreApplication::exit() to avoid removal of pid file
      ::exit(2);
    }
  }

  void DaemonApplication::writePidFile(pid_t pid)
  {
    std::string pf=pidFile(applicationName());
    FILE * fpid=fopen(pf.data(),"wt");
    if(fpid) {
      fprintf(fpid, "%u\n", pid);
      fclose(fpid);
    } else {
      Log::write(0, "cannot create %s\n", pf.data());
    }
  }

  /*!
    Checks if basepath exists and creates directory if it does not exist yet.
  */
  void DaemonApplication::checkBasePath()
  {
    struct stat sInfo;
    int sErr=stat(_basePath.data(), &sInfo);
    if(sErr!=0) {
      fprintf(stderr, "%s: %s does not exist\n", applicationName(), _basePath.data());
      exit(2);
    }
    if(!S_ISDIR(sInfo.st_mode)) {
      fprintf(stderr, "%s: %s already exists and is not a directory\n", applicationName(), _basePath.data());
      exit(2);
    }
    _basePath+="/";
  }

  void DaemonApplication::initLogFile()
  {
    _log=new Log;
    if(_doFork) {
      std::string logFile;
      if(_logFileName.size()==0) {
        logFile=_basePath;
        logFile+=applicationName();
        logFile+=".log";
        _log->setFile(logFile.data());
      } else {
        _log->setFile(_logFileName.data());
      }
    }
  }

  void DaemonApplication::showHelp(int& i, int argc, char ** argv, ApplicationHelp * ( *help) ())
  {
    ApplicationHelp * h=help();
    setHelp(h);
    if(checkOptionArg(i, argc, argv, false)) {
      h->exec(argv[i]);
    } else {
      h->exec();
    }
    delete h;
  }

  /*!

  */
  void DaemonApplication::setHelp(ApplicationHelp * h)
  {
    h->addGroup("Daemon", "daemon");
    h->addOption("-N", "Don't go into background");
    h->addOption("-b <PATH>", "Sets base path for log and data files (default=/var/log).");
    h->addOption("-l <FILE>", "Sets log file name (default=PATH/APPLICATION_NAME.log, see option '-b' for PATH).");
    h->addOption("-no-log", "Mute all logs.");
    CoreApplication::setHelp(h);
  }

  void DaemonApplication::parentInterrupted(int)
  {
    // Kill child process
    std::string pf=pidFile(CoreApplication::instance()->applicationName());
    FILE * fpid=fopen(pf.data(), "rt");
    if(fpid) {
      pid_t pid;
      if(fscanf(fpid, "%u\n", &pid)==1) {
        kill(pid, SIGINT);
      }
      fclose(fpid);
    }
    // Nicer output of runscript if parent process is printing messages
    printf("\n");
    exit(2);
  }

  void DaemonApplication::childInterrupted(int)
  {
    // Do not use TRACE here because it uses a mutex which can be locked while this function
    // is called inside a signal exception.
    Log::writeNoMutex(0, "received TERM or INTERRUPT signal\n");
    EventLoop::mainInstance()->terminate();
    // Returns to default handler in case the process did not stop properly
    signal(SIGTERM, SIG_DFL);
    signal(SIGINT, SIG_DFL);
  }

  void DaemonApplication::installInterruptSignals()
  {
    signal(SIGTERM, childInterrupted);
    signal(SIGINT, childInterrupted);
  }

  void DaemonApplication::fork()
  {
    if(_doFork) {
      // Our process ID and Session ID
      pid_t pid, sid;
      pid=::fork();
      if(pid<0) { // Error, no child is created, -1 is returned to parent
        Log::write(0, "cannot fork off the parent process: %s\n", strerror(errno));
        exit(2);
      }
      // If we got a good PID, then we can exit the parent process.
      if(pid>0) {
        writePidFile(pid);
        // Eventually wait before exiting
        signal(SIGTERM, parentInterrupted);
        signal(SIGINT, parentInterrupted);
        parentProcessWait();
        ::exit(0); // Exits silently
      }
      // pid is null, hence we are in the child process
      // Change the file mode mask
      umask(0);
      // Create a new SID for the child process
      sid=setsid();
      if(sid < 0) {
        Log::write(0, "cannot create a new SID for the child process\n");
        exit(2);
      }
      initInternalDebugger();
    }
  }

  void DaemonApplication::start()
  {
    installInterruptSignals();

    Log::write(0, "start new daemon with pid %i\n",getpid());
  }

} // namespace GpCoreTools
