/***************************************************************************
**
**  This file is part of GpCoreTools.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program. If not, see <http://www.gnu.org/licenses/>.
**
**  See http://www.geopsy.org for more information.
**
**  Created: 2009-08-19
**  Authors:
**    Marc Wathelet
**    Marc Wathelet (LGIT, Grenoble, France)
**
***************************************************************************/

#include <poll.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "EventLoop.h"
#include "Stream.h"
#include "Timer.h"
#include "PThread.h"
#include "Log.h"
#include "EventStream.h"
#include "NewTimerEvent.h"
#include "Trace.h"

namespace GpCoreTools {

  /*!
    \class EventLoop EventLoop.h
    \brief Brief description of class still missing

    Full description of class still missing
  */

  EventLoop * EventLoop::_mainLoop=0;

  /*!
    Description of constructor still missing
  */
  EventLoop::EventLoop()
  {
    TRACE;
    PThread * t=PThread::currentThread();
    if(t) {
      t->setEventLoop(this);
    } else {
      assert(!_mainLoop);
      _mainLoop=this;
    }
    // A way to force poll() to exit immediately
    _eventStream=new EventStream;
    addStream(_eventStream);
    _terminate=false;
  }

  /*!
    Description of destructor still missing
  */
  EventLoop::~EventLoop()
  {
    TRACE;
    for(int i=_streams.size()-1;i>=0;i--) {
      delete _streams[i];
    }
    for(std::list<Timer::QueueItem *>::iterator it=_timers.begin(); it!=_timers.end(); it++) {
      delete *it;
    }
  }

  EventLoop * EventLoop::instance()
  {
    PThread * t=PThread::currentThread();
    if(t) {
      return t->eventLoop();
    } else {
      return _mainLoop;
    }
  }

  Timer::QueueItem * EventLoop::addTimer(Timer * t)
  {
    Timer::QueueItem * tq=new Timer::QueueItem(t);
    addTimer(tq);
    // Force poll() to exit immediately
    _eventStream->send(new NewTimerEvent);
    return tq;
  }

  void EventLoop::addTimer(Timer::QueueItem * tq)
  {
    TRACE;
    Timer * t=tq->timer();
    t->setTimeout();
    std::list<Timer::QueueItem *>::iterator it;
    for(it=_timers.begin(); it!=_timers.end(); it++) {
      if((*it)->timer() && t->after((*it)->timer())) {
        break;
      }
    }
    if(it==_timers.end()) {
      _timers.push_back(tq);
    } else {
      _timers.insert(it, tq);
    }
    /*for(std::list<Timer::QueueItem *>::iterator it=_timers.begin(); it!=_timers.end(); it++) {
      (*it)->timer()->printDebug();
    }*/
  }

  void EventLoop::terminate()
  {
    // Do not use TRACE here because it uses a mutex which can be locked while this function
    // might be called inside a signal exception.
    _terminate=true;
    // Force poll() to exit immediately
    _eventStream->send(new NewTimerEvent);
  }

  /*!
    Returns true if \a maxTime is reached.
  */
  bool EventLoop::exec(int maxTime)
  {
    TRACE;
    _terminate=false;
    Log::write(1, "start event loop\n");
    // Init pollfd vector
    pollfd * pfds=(pollfd *)malloc((_streams.size())*sizeof(pollfd));
    for(int i=_streams.size()-1;i>=0;i--) {
      pollfd& pfd=pfds[i];
      Stream& s=*_streams.at(i);
      pfd.fd=s.fileDescriptor();
      pfd.events=s.eventTypes();
      //Log::write(8, "create stream %i events %x (POLLIN=%i, POLLPRI=%i)\n", i, pfd.events, POLLIN, POLLPRI);
    }
    // Main loop
    Timer loopTimer;
    if(maxTime>0) {
      loopTimer.setInterval(maxTime);
      loopTimer.start();
    }
    while(!_terminate) {
      // Calculate next timeout
      Timer::QueueItem * nextTimer=0;
      int timeout;
      if(_timers.size()==0) {
        timeout=-1;
      } else {
        while(!nextTimer && _timers.size()>0) {
          nextTimer=_timers.back();
          if(!nextTimer->timer()) { // Get rid of stopped timers
            _timers.pop_back();
            delete nextTimer;
            nextTimer=0;
          }
        }
        if(nextTimer) {
          timeout=nextTimer->timer()->timeout();
        } else {
          timeout=-1;
        }
      }
      // Wait for next event or timeout
      unsigned int n=_streams.size();
      //Log::write(6, "enter poll %i streams timeout %i\n", n, timeout);
      int nfd=poll(pfds, n, timeout);
      //Log::write(6, "exit poll %i\n", nfd);
      if(nfd==-1) {
        if(errno!=EINTR) {
          Log::write(0, "%s\n", strerror(errno));
        }
        Log::write(1, "stop event loop\n");
        return false;
      }
      // Process next timer
      if(nextTimer && nextTimer->timer()) {
        if(nextTimer->timer()->timeout()==0) {
          _timers.pop_back();
          // TODO: strange 2nd condition, check its validity
          if(nextTimer->timer()->exec() && nextTimer->timer()) {
            if(nextTimer->timer()==&loopTimer) {
              //Log::write(6, "exit loop, time out\n");
              return true;
            }
            addTimer(nextTimer);
          } else {
            delete nextTimer;
          }
        }
      }
      // Process data exchange on fds
      if(nfd>0) {
        for(unsigned int i=0;i<n;i++) {
          _streams.at(i)->event(pfds[i].revents);
        }
      }
      // Process event queue
      while(!_events.empty()) {
        Event * event=_events.front();
        event->exec();
        delete event;
        _events.pop();
      }
      // Update pollfd vector for new streams
      if(_streams.size()>n) {
        unsigned int newN=_streams.size();
        pfds=(pollfd *)realloc(pfds, newN*sizeof(pollfd));
        for(unsigned int i=n;i<newN;i++) {
          Stream& s=*_streams.at(i);
          pollfd& pfd=pfds[i];
          pfd.fd=s.fileDescriptor();
          pfd.events=s.eventTypes();
          //Log::write(8, "enlarge stream %i events %x (POLLIN=%i, POLLPRI=%i)\n", i, pfd.events, POLLIN, POLLPRI);
        }
        n=newN;
      }
      // Remove dead streams
      for(unsigned int i=0;i<n;) {
        if(_streams.at(i)->isActive()) {
          i++;
        } else {
          n--;
          delete _streams.at(i);
          pfds[i].fd=pfds[n].fd;
          pfds[i].events=pfds[n].events;
          _streams.at(i)=_streams.at(n);
          //Log::write(8, "remove stream %i events %x (POLLIN=%i, POLLPRI=%i)\n", i, pfds[i].events, POLLIN, POLLPRI);
        }
      }
      if(n<_streams.size()) _streams.resize(n);
    }
    Log::write(1, "stop event loop\n");
    return false;
  }

} // namespace GpCoreTools
