/***************************************************************************
**
**  This file is part of GpCoreTools.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program. If not, see <http://www.gnu.org/licenses/>.
**
**  See http://www.geopsy.org for more information.
**
**  Created: 2009-07-02
**  Authors:
**    Marc Wathelet
**    Marc Wathelet (LGIT, Grenoble, France)
**
***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include "File.h"
#include "Log.h"

namespace GpCoreTools {

  /*!
    \class File File.h
    \brief File manipulation tools

    Full description of class still missing
  */

  /*!
    Returns the line read from \a f in a dynamic string buffer \a buf of initial length of \a bufLen..
    If the line read from \a f is longer than initial \a bufLen, buffer is re-allocated and expanded.
    \a bufLen contains the final length. If \a delEOL is true remaining end of line characters
    (LF and/or CR) at the end are removed.
  */
  bool File::readLine(char *& buf, int& bufLen, FILE * f, bool delEOL)
  {
    buf[0]='\0';
    if(!fgets (buf, bufLen, f)) {
      return false;
    }
    while((int)strlen(buf)==bufLen-1) {
      int nbufLen=bufLen << 1;
      char * nbuf=new char [nbufLen];
      memmove(nbuf,buf,bufLen-1);
      delete [] buf;
      buf=nbuf;
      if(!fgets(buf+bufLen-1, bufLen+1, f)) {
        break;
      }
      bufLen=nbufLen;
    }
    if(delEOL) {
      int n=strlen(buf)-1;
      if(buf[n]=='\n') {
        if(buf[n-1]=='\r') {
          buf[n-1]='\0';
        } else {
          buf[n]='\0';
        }
      }
    }
    return true;
  }

  /*!
    For instance, \a beginFilePath can be "/dev/serial/by-id/usb-LEAS_DASCUBE".
    "/dev/serial/by-id/usb-LEAS_DASCUBE_DS011105-if00-port0" is returned, if this file exists.

    A pointer to a vector is returned. In all cases, you must delete this pointer.
  */
  std::vector<std::string> * File::complete(const std::string& beginFilePath)
  {
    std::vector<std::string> * files=new std::vector<std::string>;

    // Breaks path and file base name
    const char * baseName=strrchr(beginFilePath.data(), '/');
    int pathLength=baseName-beginFilePath.data();
    char * pathTemp=new char[pathLength+2];
    strncpy(pathTemp, beginFilePath.data(), pathLength);
    pathTemp[pathLength]='/';
    pathTemp[pathLength+1]='\0';
    std::string path(pathTemp);
    delete [] pathTemp;
    baseName++;
    int baseNameLength=strlen(baseName);

    struct dirent *dirp;
    DIR *dp;
    if(!(dp=opendir(path.data()))) {
      return files;
    }

    while((dirp=readdir(dp))) {
      if(strncmp(baseName, dirp->d_name, baseNameLength)==0) {
        files->push_back(path+std::string(dirp->d_name));
      }
    }
    closedir(dp);
    return files;
  }

  /*!
    Returns true if \a buffer at \a bytesRead index starts with \a string.
  */
  bool File::match(const char * buffer, int& bytesRead, int bytesCount, const char * string)
  {
    int n=bytesCount-bytesRead;
    int sn=strlen(string);
    if(n>=sn && strncmp(buffer+bytesRead, string, sn)==0) {
      bytesRead+=sn;
      return true;
    } else {
      return false;
    }
  }

  /*!
    Reads all file and returns its contents in a newly allocated buffer. It returns 0 if an error occured or
    if file is empty.
  */
  char * File::readAll(const char * fileName)
  {
    struct stat sInfo;
    int sErr=stat(fileName, &sInfo);
    if(sErr==0) {
      unsigned int fileLen=sInfo.st_size;
      if(fileLen>0) {
        FILE * f=fopen(fileName,"rb");
        if(f) {
          char * buf=new char [fileLen+1];
          if(fread(buf,sizeof(char),fileLen,f)==fileLen) {
            buf[fileLen]='\0';
          } else {
            delete [] buf;
            buf=0;
          }
          fclose(f);
          return buf;
        }
      }
    }
    return 0;
  }

  /*!
    Returns -1 if value is truncated and \a bytesRead is not modified.
  */
  int File::readInteger(const char * buffer, int& bytesRead, int bytesCount, bool& ok)
  {
    int i;
    int value;
    for(i=bytesRead; i<bytesCount && !isspace(buffer[i]); i++) {}
    int n=i-bytesRead;
    if(n>0 && i<bytesCount) {
      char * temp=new char[n+1];
      strncpy(temp, buffer+bytesRead, n);
      temp[n]='\0';
      value=atoi(temp);
      bytesRead=i;
      ok=true;
      delete [] temp;
    } else {
      value=-1;
      ok=false;
    }
    return value;
  }

  /*!
    Returns -1 if value is truncated and \a bytesRead is not modified.
  */
  long int File::readLongInteger(const char * buffer, int& bytesRead, int bytesCount, bool& ok)
  {
    int i;
    long int value;
    for(i=bytesRead; i<bytesCount && !isspace(buffer[i]); i++) {}
    int n=i-bytesRead;
    if(n>0 && i<bytesCount) {
      char * temp=new char[n+1];
      strncpy(temp, buffer+bytesRead, n);
      temp[n]='\0';
      value=atol(temp);
      bytesRead=i;
      ok=true;
      delete [] temp;
    } else {
      value=-1;
      ok=false;
    }
    return value;
  }

  /*!
    Returns -1 if value is truncated and \a bytesRead is not modified.
  */
  double File::readDouble(const char * buffer, int& bytesRead, int bytesCount, bool& ok)
  {
    int i;
    double value;
    for(i=bytesRead; i<bytesCount && !isspace(buffer[i]); i++) {}
    int n=i-bytesRead;
    if(n>0 && i<bytesCount) {
      char * temp=new char[n+1];
      strncpy(temp, buffer+bytesRead, n);
      temp[n]='\0';
      value=atof(temp);
      bytesRead=i;
      ok=true;
      delete [] temp;
    } else {
      value=-1;
      ok=false;
    }
    return value;
  }

} // namespace GpCoreTools
