/***************************************************************************
**
**  This file is part of GpCoreTools.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program. If not, see <http://www.gnu.org/licenses/>.
**
**  See http://www.geopsy.org for more information.
**
**  Created: 2011-04-21
**  Authors:
**    Marc Wathelet (LGIT, Grenoble, France)
**
***************************************************************************/

#include <math.h>

#include "LinearRegression.h"
#include "Log.h"

namespace GpCoreTools {

  /*!
    \class LinearRegression LinearRegression.h
    \brief Brief description of class still missing

    Full description of class still missing
  */

  /*!
    If \a n is greater than 0, keep last \a n values in the stack.
  */
  void LinearRegression::reset(int n)
  {
    if(n>0) {
      int i0=_values->size()-n;
      if(i0>0) {
        _sumX=0.0;
        _sumY=0.0;
        _sumXY=0.0;
        _sumX2=0.0;
        std::vector<std::pair<double, double> > * values=_values;
        _values=new std::vector<std::pair<double, double> >;
        for(int i=0; i<n; i++) {
          std::pair<double, double> v=(*values)[i0+i];
          add(v.first, v.second);
        }
        delete values;
        calculate(); // update a,b
      }
    } else {
      _sumX=0.0;
      _sumY=0.0;
      _sumXY=0.0;
      _sumX2=0.0;
      _a=0.0;
      _b=0.0;
      if(_values) {
        _values->clear();
      } else {
        _values=new std::vector<std::pair<double, double> >;
      }
    }
  }

  void LinearRegression::add(double x, double y)
  {
    if(_values->size()==0) { // Avoid big numbers on x
      _x0=x;
      x=0.0;
    } else {
      x-=_x0;
    }
    _sumX+=x;
    _sumY+=y;
    _sumXY+=x*y;
    _sumX2+=x*x;
    _values->push_back(std::pair<double, double>(x, y));
  }

  /*!
    Computes a and b. Returns false if line is vertical. In this case, equation is x=b.
  */
  bool LinearRegression::calculate()
  {
    double invn=1.0/_values->size();
    double denom=_sumX2-invn*_sumX*_sumX;
    if(denom!=0.0) {
      _a=(_sumXY-invn*_sumY*_sumX)/denom;
      _b=(invn*_sumY*_sumX2-invn*_sumXY*_sumX)/denom;
      return true;
    } else {
      _a=0.0;
      _b=_sumX*invn;
      return false;
    }
  }

  /*!
    Return the standard deviation around regression line.
    A and B must be computed.
  */
  float LinearRegression::stddev() const
  {
    double sum=0;
    double sum2=0;
    std::vector<std::pair<double, double> >::const_iterator it;
    for(it=_values->begin(); it!=_values->end(); it++) {
      double y=_a*it->first+_b;  // Already corrected by x0
      double dy=y-it->second;
      sum+=dy;
      sum2+=dy*dy;
    }
    double n=_values->size();
    double mean=sum/n;
    double stddev=(sum2-mean*mean*n)/(n-1);
    if(stddev<0.0) {
      return 0.0;
    } else {
      return sqrt(stddev);
    }
  }

  /*!
    Never returns an index greater that half of the size. It avoids
    rejecting values that have been just added.
  */
  int LinearRegression::badValue(double deviation) const
  {
    int n=_values->size() >> 1;
    for(int i=0; i<n; i++) {
      const std::pair<double, double>& v=(*_values)[i];
      double y=_a*v.first+_b;  // Already corrected by x0
      if(fabs(y-v.second)>deviation) {
        return i;
      }
    }
    return -1;
  }

  /*!
    Remove value at index \a i. The order of value is altered. do not use
    at the same time as removeFirst().
  */
  void LinearRegression::remove(int i)
  {
    unsigned int n=_values->size();
    std::pair<double, double>& v=(*_values)[i];
    double x=v.first;
    double y=v.second;
    _sumX-=x;
    _sumY-=y;
    _sumXY-=x*y;
    _sumX2-=x*x;
    v=(*_values)[n-1];
    _values->resize(n-1);
  }

  /*!
    For a moving estimation of the linear regression, check the size
    after each add(), and eventually remove the fisrt value.

    Do not use remove() with this function. remove() alters the order
    of values.
  */
  void LinearRegression::removeFirst()
  {
    std::pair<double, double>& v=_values->front();
    double x=v.first;
    double y=v.second;
    _sumX-=x;
    _sumY-=y;
    _sumXY-=x*y;
    _sumX2-=x*x;
    _values->erase(_values->begin());
  }

} // namespace GpCoreTools
