/***************************************************************************
**
**  This file is part of GpCoreTools.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program. If not, see <http://www.gnu.org/licenses/>.
**
**  See http://www.geopsy.org for more information.
**
**  Created: 2009-06-26
**  Authors:
**    Marc Wathelet
**    Marc Wathelet (LGIT, Grenoble, France)
**
***************************************************************************/

#ifndef MESSAGECLASSIDHEADER_H
#define MESSAGECLASSIDHEADER_H

#include "ByteOrder.h"
#include "Log.h"
#include "FletcherChecksum.h"
#include "GpCoreToolsDLLExport.h"

namespace GpCoreTools {

#pragma pack(1)     /* set alignment to 1 byte boundary */

  class GPCORETOOLS_EXPORT MessageClassIdHeader
  {
  public:
    MessageClassIdHeader() {}
    inline MessageClassIdHeader(unsigned char msgClass, unsigned char msgId, unsigned int length);
    inline MessageClassIdHeader(const MessageClassIdHeader& o);

    enum {SynchronizationSize=2, HeaderSize=4, CheckSumSize=2};

    bool isValid() const {return _sync[0]==0xB5 && _sync[1]==0x62;}
    inline bool isValid(const FletcherChecksum& cksum, const char * payload=0,
                        unsigned short payloadLength=0) const;

    unsigned char messageClass() const {return _values.messageClass;}
    unsigned char messageId() const {return _values.messageId;}

    inline void littleEndianValues();
    inline void bigEndianValues();

    unsigned int lengthLittleEndian() const {return ByteOrder::littleEndianToNative(_values.length);}
    unsigned int lengthBigEndian() const {return ByteOrder::bigEndianToNative(_values.length);}

    const char * header() const {return _bytes;}
  protected:
    unsigned char _sync[2];
    union {
      struct {
        unsigned char messageClass;
        unsigned char messageId;
        unsigned short length;
      } _values;
      char _bytes[HeaderSize];
    };
  };

  inline MessageClassIdHeader::MessageClassIdHeader(unsigned char msgClass, unsigned char msgId, unsigned int length)
  {
    _sync[0]=0xB5;
    _sync[1]=0x62;
    _values.messageClass=msgClass;
    _values.messageId=msgId;
    _values.length=length;
  }

  inline MessageClassIdHeader::MessageClassIdHeader(const MessageClassIdHeader& o)
  {
    _sync[0]=o._sync[0];
    _sync[1]=o._sync[1];
    _values.messageClass=o._values.messageClass;
    _values.messageId=o._values.messageId;
    _values.length=o._values.length;
  }

  inline void MessageClassIdHeader::bigEndianValues()
  {
#if BYTE_ORDER==LITTLE_ENDIAN
    ByteOrder::swap(_bytes[2], _bytes[3]);
#endif
  }

  inline void MessageClassIdHeader::littleEndianValues()
  {
#if BYTE_ORDER==BIG_ENDIAN
    ByteOrder::swap(_bytes[2], _bytes[3]);
#endif
  }

  inline bool MessageClassIdHeader::isValid(const FletcherChecksum& cksum, const char * payload,
                                            unsigned short payloadLength) const
  {
    FletcherChecksum mycs;
    mycs.add(reinterpret_cast<const char *>(header()), HeaderSize);
    if(payload) {
      mycs.add(payload, payloadLength);
    }
    if(cksum==mycs) {
      return true;
    } else {
      Log::write(5, "bad checksum 0x%02hhX 0x%02hhX (in message) instead of 0x%02hhX 0x%02hhX (computed)\n", cksum.a(), cksum.b(), mycs.a(), mycs.b());
      return false;
    }
  }

#pragma pack()   // restore original alignment

} // namespace GpCoreTools

#endif // MESSAGECLASSIDHEADER_H
