/***************************************************************************
**
**  This file is part of GpCoreTools.
**
**  This file may be distributed and/or modified under the terms of the
**  GNU General Public License version 2 or 3 as published by the Free
**  Software Foundation and appearing in the file LICENSE.GPL included
**  in the packaging of this file.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
**  more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program. If not, see <http://www.gnu.org/licenses/>.
**
**  See http://www.geopsy.org for more information.
**
**  Created: 2009-06-26
**  Authors:
**    Marc Wathelet
**    Marc Wathelet (LGIT, Grenoble, France)
**
***************************************************************************/

#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <stdio.h>
#include <poll.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "Serial.h"
#include "EventLoop.h"
#include "Log.h"
#include "Trace.h"

namespace GpCoreTools {

  /*!
    \class Serial Serial.h
    \brief Brief description of class still missing

    Full description of class still missing
  */

  /*!
    Description of constructor still missing

    \a buffer can be null for single way serial connection
  */
  Serial::Serial(const std::string& devicePath, speed_t speed, tcflag_t flags, DynamicBuffer * buffer)
  {
    TRACE;
    _speed=speed;
    _flags=flags;
    _devicePath=devicePath;
    _buffer=buffer;
    assert(pthread_mutex_init(&_mutex, 0)==0);
  }

  /*!
    Description of destructor still missing
  */
  Serial::~Serial()
  {
    TRACE;
    delete _buffer;
  }

  void Serial::hungUp()
  {
    TRACE;
    setFileDescriptor(-1);
    connect();
  }

  bool Serial::isConnected()
  {
    TRACE;
    lock();
    bool a=isActive();
    unlock();
    return a;
  }

  /*!
    Returns false if connection times out and if event loop is interrupted.
  */
  bool Serial::connect()
  {
    TRACE;
    lock();
    while(!open()) {
      unlock();
      sleep(5);
      if (EventLoop::instance()->isTerminated()) {
        return false;
      }
      lock();
    }
    init();
    unlock();
    Log::write(0, "%s connected\n", _devicePath.data());
    return true;
  }


  /*!
    Returns the output speed from a termios structure.
    \internal
  */
  int Serial::outputSpeed(termios * t)
  {
    TRACE;
    speed_t code=cfgetospeed(t);
    switch (code) {
    case B0:     return 0;
    case B50:    return 50;
    case B75:    return 75;
    case B110:   return 110;
    case B134:   return 134;
    case B150:   return 150;
    case B200:   return 200;
    case B300:   return 300;
    case B1200:  return 1200;
    case B2400:  return 2400;
    case B4800:  return 4800;
    case B9600:  return 9600;
    case B19200: return 19200;
    case B38400: return 38400;
    case B57600: return 57600;
    default: return 115200;
    }
  }

  /*!
    Opens a serial communication with device
  */
  bool Serial::open()
  {
    TRACE;
    setFileDescriptor(::open(_devicePath.data(), O_RDWR | O_NOCTTY | O_NDELAY));
    if(fileDescriptor()<0) {
      Log::write(1, "%s: %s\n", _devicePath.data(), strerror(errno));
      return false;
    } else {
      if(_buffer) {
        _buffer->setFileDescriptor(fileDescriptor());
      }
      return true;
    }
  }

  /*!
    Initializes terminal parameters
  */
  void Serial::init()
  {
    TRACE;
    termios t;
    tcgetattr(fileDescriptor(), &t);
    //printf("Output speed %i bauds\n",outputSpeed(&t));
    cfmakeraw(&t);
    cfsetospeed(&t, _speed);
    cfsetispeed(&t, _speed);
    t.c_oflag&=~(ONLCR);
    t.c_iflag&=~(PARMRK | INPCK);
    // PARENB: (0x0100) Enable parity generation on output and parity checking for input.
    // PARODD: (0x0200) If set, then parity for input and output is odd; otherwise even  parity is used.
    // CSIZE: (0x0030) Character size mask.  Values are CS5 (0x00), CS6(0x10), CS7(0x20), or CS8(0x30)
    // Set to 0 all bits of interest
    t.c_cflag&=~(CSIZE | PARENB | PARODD);
    // Enable interesting control flags
    t.c_cflag|=_flags;

    tcsetattr(fileDescriptor(), TCSANOW, &t);
    tcflush(fileDescriptor(), TCIOFLUSH);
    usleep(200000);
    tcflush(fileDescriptor(), TCIOFLUSH);
  }

  void Serial::event(short type)
  {
    TRACE;
    if(type & (POLLIN | POLLPRI)) {
      if(type & POLLPRI) {
        Log::write(8, "priority data ready (type=%i, POLLIN=%i, POLLPRI=%i\n", type, POLLIN, POLLPRI);
      }
      if(_buffer) {
        if(!_buffer->read()) {
          Log::write(1, "error reading data (socket %i)\n", fileDescriptor());
        }
      }
    }
    if(type & (POLLRDHUP | POLLHUP)) {
      Log::write(1, "connection hung up\n");
      close();
      hungUp();
    }
  }

  /*!
    \fn bool Serial::write(const char * buf, ssize_t byteCount)
    Write data through current serial line. It returns true if all bytes
    were successfully written.
  */

  /*!
    \fn DynamicBuffer * Serial::buffer() const
    Returns current data input buffer.
  */

} // namespace GpCoreTools
